####
# Test suite for the epop_ephemeris_library.
# To run:
# >>> import epop_ephemeris_library.__tester__
# >>> epop_ephemeris_library.__tester__.test_all()


if __name__ == '__main__':
  print("Cannot run the Test-suite over CLI. Please import from a python environment.")
  print(">>> import epop_ephemeris_library.__tester__")
  print(">>> epop_ephemeris_library.__tester__.test_all()")
  exit()
  
# Standard Libraries.
import os
import sys
import datetime
import warnings
# External Libraries
import numpy
from numpy.lib import recfunctions

TEST_DT_START = datetime.datetime(2021,1,1)   # 2021-01-01 00:00:00 UTC
TEST_DT_END   = datetime.datetime(2021,1,1,1) # 2021-01-01 01:00:00 UTC

MET_EPOCH = datetime.datetime(1968,5,24)

##########
def test_all():
  test_attitude()
  test_common()
  test_eclipse()
  test_ephemeris()
  test_quaternions()
  test_frame_translation()
  test_epop_geopack()
  
##########
def test_attitude():
  '''
  Tests the attitude library for runtime errors.
  '''
  from epop_ephemeris_library import attitude as attitude
  # Test various fetches
  try:
    basic_data    = attitude.get_yaw_pitch_roll       ( TEST_DT_START, TEST_DT_END, minimum_quality=None, detail=False )
  except Exception as E:
    raise ValueError("Basic fetch of YPR data failed.") from E

  try:
    basic_details = attitude.get_yaw_pitch_roll( TEST_DT_START, TEST_DT_END, minimum_quality=None, detail=True)
  except Exception as E:
    raise ValueError("Detailed fetch of YPR data failed.") from E
  
  # Fetch some with various sub-args
  _qual_name = ""
  try:
    for qual_name in attitude.common.QUALITY_FLAG_STR_SHORT:
      _qual_name = qual_name
      attitude.get_yaw_pitch_roll( TEST_DT_START, TEST_DT_END, minimum_quality=qual_name, detail=False)
      attitude.get_yaw_pitch_roll( TEST_DT_START, TEST_DT_END, minimum_quality=qual_name, detail=True )
  except Exception as E:
    raise ValueError("Error in fetching YPR data using min_quality "+_qual_name) from E   
   
  if basic_data.shape != basic_details.shape:
    raise ValueError( "Attitude Basic vs Detail Shape mismiatch." )
  
  
  _ = attitude.interpolate_ypr( basic_details, basic_details['met'][1:-1], ignore_dropout=False )
  _ = attitude.interpolate_ypr( basic_details, basic_details['met'][1:-1], ignore_dropout=True  )
  _ = attitude.interpolate_ypr( basic_details, basic_details['met']+5,     ignore_dropout=False ) # Extrapolation tests.  
  _ = attitude.interpolate_ypr( basic_details, basic_details['met']-5,     ignore_dropout=False )  
    
def test_common():
  '''
  Tests the Common library for runtime errors.
  '''
  from epop_ephemeris_library import common as common
  
  # To test:
  # add_utc_column( data: numpy.array, met_column_name: str = 'met', utc_column_name: str = 'utc'):
  # add_string_column( data: numpy.array, index_column_name: str, output_column_name: str, index_lookup_table: dict ):
  
  # _check_version()
  # get_version()
  
  # get_wgs84_earth_radius_meters(x: float, y: float, z: float) -> float:
  
  # utc_to_met(date_time: Union[datetime.datetime, str]) -> float:
  # met_to_utc(met: float) -> datetime.datetime:
  # utc2met(date_time: Union[datetime.datetime, str]) -> float:
  # met2utc(met: float) -> datetime.datetime:
 
  # Test the time-translation functions.
  _Start_MET = common.utc_to_met( TEST_DT_START ) # Also acts as the test-harness for the translation functions.
  _End_MET   = common.utc_to_met( TEST_DT_END   )
  
  _ = common.utc_to_met( "2021-01-01T00:00:00" ) # Test the string-comprehension.  
  
  if TEST_DT_START != common.met_to_utc( _Start_MET ):
    raise ValueError(" met_to_utc does not invert utc_to_met!")
  if TEST_DT_END != common.met_to_utc( _End_MET ):
    raise ValueError(" met_to_utc does not invert utc_to_met!")
    

  # Build an example dataset to add strings & UTC to:
  MET = numpy.arange( _Start_MET, _End_MET )
  # Use the common.ATTITUDE_DATA_SOURCE_STRING dict as the input dataset
  _minVal = min( common.ATTITUDE_DATA_SOURCE_STRING )
  _maxVal = max( common.ATTITUDE_DATA_SOURCE_STRING)
  Rand_Ints = numpy.random.randint( _minVal, _maxVal+1, len(MET) )
  # Build the structured array as needed:
  _data = numpy.column_stack(( MET, Rand_Ints ))
  _dtype = numpy.dtype( [('met',    'float64'),
                         ('in_test','int32')] )
  Data = recfunctions.unstructured_to_structured( _data, dtype=_dtype)
  
  # Add the 'utc' and 
  Data = common.add_utc_column( Data, 'met', 'utc' )
  Data = common.add_string_column( Data, 'in_test', 'out_test', common.ATTITUDE_DATA_SOURCE_STRING )
  
  # Can't run test_version without pinging the database, but it is called on each query of the database.
  # Assume it's valid, as it'll break on all queries if it fails.
  # Can test the 'get_version' though!
  _ = common.get_version()
  
  # Test the WGS84 conversion
  _rand_x = numpy.random.rand(100) * 1e7
  _rand_y = numpy.random.rand(100) * 1e7
  _rand_z = numpy.random.rand(100) * 1e7
  _ = common.get_wgs84_earth_radius_meters( _rand_x, _rand_y, _rand_z)
    
def test_eclipse():
  '''
  Tests the eclipse library for runtime errors.
  '''
  from epop_ephemeris_library import eclipse as eclipse
  One_Hz         = eclipse.get_eclipse( TEST_DT_START, TEST_DT_END, range_times=False)
  Eclipse_Ranges = eclipse.get_eclipse( TEST_DT_START, TEST_DT_END, range_times=True )
  # And test the interpolation:
  _ = eclipse.interpolate_eclipse(One_Hz, One_Hz['met']     )
  _ = eclipse.interpolate_eclipse(One_Hz, One_Hz['met'][::2])
  _ = eclipse.interpolate_eclipse(One_Hz, One_Hz['met']+5   )
  
def test_ephemeris():
  '''
  Tests the ephemeris library for runtime errors.
  '''
  from epop_ephemeris_library import ephemeris as ephemeris
  GEO = ephemeris.get_geo_ephemeris( TEST_DT_START, TEST_DT_END, position=True, velocity=True )
  J2K = ephemeris.get_j2k_ephemeris( TEST_DT_START, TEST_DT_END, position=True, velocity=True )
  
  # And test the variants.
  _ = ephemeris.get_geo_ephemeris( TEST_DT_START, TEST_DT_END, position=False, velocity=True )
  _ = ephemeris.get_j2k_ephemeris( TEST_DT_START, TEST_DT_END, position=False, velocity=True )
  _ = ephemeris.get_geo_ephemeris( TEST_DT_START, TEST_DT_END, position=True, velocity=False )
  _ = ephemeris.get_j2k_ephemeris( TEST_DT_START, TEST_DT_END, position=True, velocity=False )

  # And test the get_predicted.
  # Get the predicted Ranges:
  _Pred_Start, _Pred_Stop = ephemeris.get_ephemeris_range( get_definitive = False, get_predicted = True, return_met = True )
  _ = ephemeris.get_predicted_ephemeris( _Pred_Start, _Pred_Start + 600, True, True) # Get 10m of data.  
  
  if GEO.shape != J2K.shape:
    print("Size mismatch in the GEO & J2K ephemeris data. Should be 1Hz!")
  
  GEO_w_alt = ephemeris.add_altitude_column( GEO, altitude_above_earth= False )
  J2K_w_alt = ephemeris.add_altitude_column( J2K, altitude_above_earth= False )
  _         = ephemeris.add_altitude_column( GEO, altitude_above_earth= True ) # Test Alt-above earth.
   
  _ = ephemeris.add_apogee_perigee_column( GEO_w_alt ) # Test apogee, perigee list.
  _ = ephemeris.add_apogee_perigee_column( J2K_w_alt )
  
  _ = ephemeris.interpolate_ephemeris( GEO, GEO['met'][1:-1]) # Test interpolation.
  _ = ephemeris.interpolate_ephemeris( GEO, GEO['met'] + 5  ) # Test extrapolation.
  
def test_quaternions():
  '''
  Tests the quaternions library for runtime errors.
  '''
  from epop_ephemeris_library import quaternions as quaternions
  
  qTest = quaternions.get_attitude_quaternions( TEST_DT_START, TEST_DT_END, minimum_quality="No Solution", detail=True )
  _     = quaternions.get_attitude_quaternions( TEST_DT_START, TEST_DT_END, minimum_quality="No Solution", detail=False ) # Test w/o detail
  _     = quaternions.get_attitude_quaternions( TEST_DT_START, TEST_DT_END, minimum_quality=None,          detail=False ) # Test w/o min-qual arg.
  _     = quaternions.get_attitude_quaternions( TEST_DT_START, TEST_DT_END, minimum_quality="Coarse Sun Sensors", detail=False ) # Test filter.

  _ = quaternions.interpolate_quaternions( qTest, qTest['met'][1:-1], ignore_dropout=False ) # Interpolation, w/dropouts
  _ = quaternions.interpolate_quaternions( qTest, qTest['met']+10,    ignore_dropout=True  ) # Extrapolation, w/o dropouts.
  
  _ = quaternions.add_pointing_dir( qTest, 'Ram'      )
  _ = quaternions.add_pointing_dir( qTest, '-y'       )
  _ = quaternions.add_pointing_dir( qTest, 'Port'     )
  _ = quaternions.add_pointing_dir( qTest, [0,0.5,0.5])

def test_frame_translation():
  '''
  Tests the frame_translation library for runtime errors.
  '''
  from epop_ephemeris_library import ephemeris         as ephemeris
  from epop_ephemeris_library import frame_translation as frame_translation
  
  GEO = ephemeris.get_geo_ephemeris( TEST_DT_START, TEST_DT_END, position=True, velocity=False )
  J2K = ephemeris.get_j2k_ephemeris( TEST_DT_START, TEST_DT_END, position=True, velocity=True  )
  
  # Basic completion test.
  _ = frame_translation.add_frame( GEO, frame_translation.AVAILABLE_FRAMES, cartesian=True , spherical=True )
  _ = frame_translation.add_frame( GEO, frame_translation.AVAILABLE_FRAMES, cartesian=False, spherical=True )
  _ = frame_translation.add_frame( GEO, frame_translation.AVAILABLE_FRAMES, cartesian=True , spherical=False)
  # Test full names
  _ = frame_translation.add_frame( GEO, frame_translation.AVAILABLE_FRAMES_BY_NAME, cartesian=True , spherical=True )
  
  # Test overwriting.
  _tmp = frame_translation.add_frame( GEO,  frame_translation.AVAILABLE_FRAMES, cartesian=True , spherical=True )
  _tmp = frame_translation.add_frame( _tmp, frame_translation.AVAILABLE_FRAMES, cartesian=True , spherical=True )
  
  # Test Solar-Vectors.
  _ = frame_translation.add_nadir_solar_angles( J2K )
  # Test overwriting/chaining.
  _tmp = frame_translation.add_nadir_solar_angles( J2K  )
  _tmp = frame_translation.add_nadir_solar_angles( _tmp )
  
  # Test J2K->GEO
  _ = frame_translation._add_frame_geo( J2K )
  
  

def test_epop_geopack( test_dt = None ):
  '''
  Tests the epop_geopack library for accuracy.
  Tests all available internal functions.  
  Requires the 'geopack' library not used anywhere else in the package.
  '''
  # Check that the 'geopack' library is installed, as it's required to do the following tests.
  try:
    import geopack as Old_Geopack
  except:
    warnings.warn("Could not import the 'geopack' library to compare the new epop_geopack library against. Please install this non-required library to test this portion of the library.")
    return
  
  from epop_ephemeris_library.rotation_libraries import lib_epop_geopack as New_Geopack
    
  # If the differece in values exceed either the angle or the vector-length (Altitude), then throw warnings.
  ANGLE_LIMIT  = 1e-6 # Radians.
  LENGTH_LIMIT = 0.1
  
  
  if test_dt is None:
    # Get a random  of Unix-timestamps for each year valid in Geopack.
    _test_years     = numpy.arange(1895,2031)
    _test_unix      = numpy.array([( datetime.datetime(Year,1,1) - datetime.datetime(1970,1,1) ).total_seconds() for Year in _test_years])
    test_unix_times = _test_unix + numpy.random.rand( len(_test_unix) ) * 365*86400 # Sub-microsecond accuracy. Doesn't matter, as all equations are keyed on floats.
  else:
    # Cast to list if given singleton.
    if isinstance( test_dt, datetime.datetime ):
      test_dt = [test_dt]
    # And then make into unix-list.
    test_unix_times = numpy.array([( utc_dt - datetime.datetime(1970,1,1) ).total_seconds() for utc_dt in test_dt])

  # Generate some random ephemeris points.
  # Note that the position is frameless. Each frame is given indpendantly as to avoid issues with chaining.
  # Generate random cartesian coordinates - Nx3 set of cartesian points at about Cassiope's altittude.

  rand_cart_pos = numpy.random.rand( len(test_unix_times), 3 ) * 1e7 
  # And make sure all altitudes are above earth's surface.
  rand_cart_pos[ numpy.sum( rand_cart_pos**2, axis=1 )**0.5 < 6.378e6] += 6.4e6 * 2 # x2 so can ignore the sign, as guarnanteed -6.378:+6.378

  # Manually generate the Spherical Coordinates.
  # Radius, Theta Phi. (Altitude, Longitude, Latitude). Angles in Rad.
  _Radius = numpy.sum( rand_cart_pos**2, axis=1 )**0.5
  _Theta  = numpy.arctan2( rand_cart_pos[:,1], rand_cart_pos[:,0] ) # Longitude = arctan2( y, x)
  _Phi    = numpy.arccos ( rand_cart_pos[:,2] / _Radius )
  rand_sph_pos = numpy.column_stack(( _Radius, _Theta, _Phi ))

  Geographic_Radius     = rand_sph_pos[:,0] / 1000 # Kilometers from center of earth.
  Geographic_Latitude   = rand_sph_pos[:,1]
  Geographic_Colatitude = numpy.pi/2 - Geographic_Latitude

  print("Testing epop_geopack...")
  print("Failure Tolerances set at:")
  print("Angle:    {:s} Rad = {:.3f} arcseconds".format( str(ANGLE_LIMIT), ANGLE_LIMIT * 180/numpy.pi * 3600) )
  print("Position: {:.2f} meters".format( LENGTH_LIMIT ))
  print()
  
  # Init the bulk-recalc values for epopGeopack.
  New_Geopack.recalc( test_unix_times )
  # Test each function individually.
  _New_gsw_to_gsm  = New_Geopack.gswgsm ( *rand_cart_pos.T,  1 )
  _New_gsm_to_gsw  = New_Geopack.gswgsm ( *rand_cart_pos.T, -1 )
  _New_geo_to_mag  = New_Geopack.geomag ( *rand_cart_pos.T,  1 )
  _New_mag_to_geo  = New_Geopack.geomag ( *rand_cart_pos.T, -1 )
  _New_gei_to_geo  = New_Geopack.geigeo ( *rand_cart_pos.T,  1 )
  _New_geo_to_gei  = New_Geopack.geigeo ( *rand_cart_pos.T, -1 )
  _New_mag_to_sm   = New_Geopack.magsm  ( *rand_cart_pos.T,  1 )
  _New_sm_to_mag   = New_Geopack.magsm  ( *rand_cart_pos.T, -1 )
  _New_gsm_to_gse  = New_Geopack.gsmgse ( *rand_cart_pos.T,  1 )
  _New_gse_to_gsm  = New_Geopack.gsmgse ( *rand_cart_pos.T, -1 )
  _New_sm_to_gsm   = New_Geopack.smgsm  ( *rand_cart_pos.T,  1 )
  _New_gsm_to_sm   = New_Geopack.smgsm  ( *rand_cart_pos.T, -1 )
  _New_geo_to_gsm  = New_Geopack.geogsm ( *rand_cart_pos.T,  1 )
  _New_gsm_to_geo  = New_Geopack.geogsm ( *rand_cart_pos.T, -1 )
  
  _New_car_to_sph  = New_Geopack.sphcar ( *rand_cart_pos.T, -1 )
  _New_sph_to_car  = New_Geopack.sphcar ( *rand_sph_pos .T,  1 )

  # As altitude guarnatneed above Earth Radius, can use for both Geographic & Geodetic altitude.
  _New_geod_to_geo = New_Geopack.geodgeo( Geographic_Radius, Geographic_Latitude,    1 ) # < H, xmu  > h=Geodetic Lat in km, xmu=Geodetic latitude.
  _New_geo_to_geod = New_Geopack.geodgeo( Geographic_Radius, Geographic_Colatitude, -1 ) # < r, Theta> R=Geocentric KM, Theta = Spherical Co-latitude.
  
  #_New_geo_to_geod -> 
  _Geodetic_Altitude = _New_geo_to_geod[0]
  _Geodetic_Latitude = _New_geo_to_geod[1]
  _New_geo_to_geo  = New_Geopack.geodgeo( _Geodetic_Altitude, _Geodetic_Latitude,  1 )
  
  # And for the IGRF-bound entries:
  #_New_bsp_to_car  = New_Geopack.bspcar ( *Rand_GEO_XYZ.T,  1 ) # <theta, phi, br, btheta, bphi>
  #_New_car_to_bsp  = New_Geopack.bcarsp ( *Rand_GEO_XYZ.T, -1 ) # <x, y, z, bx, by, bz>

  

  _Old_gsw_to_gsm  = []
  _Old_gsm_to_gsw  = []
  _Old_geo_to_mag  = []
  _Old_mag_to_geo  = []
  _Old_gei_to_geo  = []
  _Old_geo_to_gei  = []
  _Old_mag_to_sm   = []
  _Old_sm_to_mag   = []
  _Old_gsm_to_gse  = []
  _Old_gse_to_gsm  = []
  _Old_sm_to_gsm   = []
  _Old_gsm_to_sm   = []
  _Old_geo_to_gsm  = []
  _Old_gsm_to_geo  = []
  _Old_geod_to_geoo= []
  _Old_geoo_to_geod= []
  _Old_sph_to_car  = []
  _Old_car_to_sph  = []
  _Old_bsp_to_car  = []
  _Old_car_to_bsp  = []
  #_Old_bcar_to_sp  = []
  #_Old_sp_to_bcar  = []
  _Old_geod_to_geo = []
  _Old_geo_to_geod = []
  
  for i in range(len( test_unix_times )):
    CARx, CARy, CARz   = rand_cart_pos[i]
    Radius, Theta, Phi = rand_sph_pos[i]
    
    # Recalc for the given time.
    Old_Geopack.recalc( test_unix_times[i] )

    _Old_gsw_to_gsm.append( Old_Geopack.gswgsm (CARx, CARy, CARz,  1) )
    _Old_gsm_to_gsw.append( Old_Geopack.gswgsm (CARx, CARy, CARz, -1) )
    _Old_geo_to_mag.append( Old_Geopack.geomag (CARx, CARy, CARz,  1) )
    _Old_mag_to_geo.append( Old_Geopack.geomag (CARx, CARy, CARz, -1) )
    _Old_gei_to_geo.append( Old_Geopack.geigeo (CARx, CARy, CARz,  1) )
    _Old_geo_to_gei.append( Old_Geopack.geigeo (CARx, CARy, CARz, -1) )
    _Old_mag_to_sm .append( Old_Geopack.magsm  (CARx, CARy, CARz,  1) )
    _Old_sm_to_mag .append( Old_Geopack.magsm  (CARx, CARy, CARz, -1) )
    _Old_gsm_to_gse.append( Old_Geopack.gsmgse (CARx, CARy, CARz,  1) )
    _Old_gse_to_gsm.append( Old_Geopack.gsmgse (CARx, CARy, CARz, -1) )
    _Old_sm_to_gsm .append( Old_Geopack.smgsm  (CARx, CARy, CARz,  1) )
    _Old_gsm_to_sm .append( Old_Geopack.smgsm  (CARx, CARy, CARz, -1) )
    _Old_geo_to_gsm.append( Old_Geopack.geogsm (CARx, CARy, CARz,  1) )
    _Old_gsm_to_geo.append( Old_Geopack.geogsm (CARx, CARy, CARz, -1) )
    
    _Old_sph_to_car .append( Old_Geopack.sphcar (Radius, Theta, Phi, 1) )
    _Old_car_to_sph .append( Old_Geopack.sphcar (CARx, CARy, CARz,  -1) )
  
    # As altitude guarnatneed above Earth Radius, can use for both Geographic & Geodetic altitude.
    Colatitude       = numpy.pi/2 - Theta
    _Old_geod_to_geo.append( Old_Geopack.geodgeo( Radius/1000, Theta,       1) )
    _Old_geo_to_geod.append( Old_Geopack.geodgeo( Radius/1000, Colatitude, -1) )    
    
    #_Old_bsp_to_car .append( geopack.bspcar () )
    #_Old_bcar_to_sp .append( geopack.bcarsp () )

    
  # For each Cartesian-rotation pair, compare the angle between the given solutions, for the cartesian values.
  
  
  # Returns Angle, length difference between solutions.
  def Build_Stats( vA, vB ):
    # A.B = |A|*|B|*cos(Angle)
    # Angle = arccos( A.B / (|A|*|B|))
    len_A   = numpy.sum( vA**2 , axis=1 ) **0.5
    len_B   = numpy.sum( vB**2 , axis=1 ) **0.5
    DotProd = numpy.sum( vA*vB, axis=1) # Informal dot product.
    
    # Pre-calc angle to filter any issues with floating point 1>1.0
    _Angle_Inner = DotProd / (len_A * len_B)
    _Angle_Inner[ _Angle_Inner > 1.0 ] = 1.0
    
    Angle = numpy.arccos( _Angle_Inner )
        
    return numpy.abs(Angle), numpy.abs(len_A - len_B) # Radian, length.
  
  
  Cartesian_Data = [\
    [_New_gsw_to_gsm , _Old_gsw_to_gsm , "gsw_to_gsm "],\
    [_New_gsm_to_gsw , _Old_gsm_to_gsw , "gsm_to_gsw "],\
    [_New_geo_to_mag , _Old_geo_to_mag , "geo_to_mag "],\
    [_New_mag_to_geo , _Old_mag_to_geo , "mag_to_geo "],\
    [_New_gei_to_geo , _Old_gei_to_geo , "gei_to_geo "],\
    [_New_geo_to_gei , _Old_geo_to_gei , "geo_to_gei "],\
    [_New_mag_to_sm  , _Old_mag_to_sm  , "mag_to_sm  "],\
    [_New_sm_to_mag  , _Old_sm_to_mag  , "sm_to_mag  "],\
    [_New_gsm_to_gse , _Old_gsm_to_gse , "gsm_to_gse "],\
    [_New_gse_to_gsm , _Old_gse_to_gsm , "gse_to_gsm "],\
    [_New_sm_to_gsm  , _Old_sm_to_gsm  , "sm_to_gsm  "],\
    [_New_gsm_to_sm  , _Old_gsm_to_sm  , "gsm_to_sm  "],\
    [_New_geo_to_gsm , _Old_geo_to_gsm , "geo_to_gsm "],\
    [_New_gsm_to_geo , _Old_gsm_to_geo , "gsm_to_geo "],\
    [_New_sph_to_car , _Old_sph_to_car , "sph_to_car "]]

  #return Cartesian_Data
  
  for i in range(len(Cartesian_Data)):
    New_Method, Old_Method, Name = Cartesian_Data[i]
    # Translate into sane formats..
    New_Method = numpy.column_stack(New_Method) # Returend as 3, Nx1 arrays, so combine into 1 Nx3
    Old_Method = numpy.array(Old_Method)        # Just need to cast to numpy.array to make calc easier.
    
    
                  
    AngleDelta, LenDelta = Build_Stats( New_Method, Old_Method )
    
    if numpy.any(AngleDelta > ANGLE_LIMIT) or numpy.any(LenDelta > LENGTH_LIMIT):
      print("FAIL:", Name)
      
      # And then filter for the issues:
      #for j in range(len(New_Method)):
      #  if not( AngleDelta[j] > ANGLE_LIMIT or LenDelta[j] > LENGTH_LIMIT): continue
      #  # Issue found:
      #  DT = datetime.datetime(1970,1,1) + datetime.timedelta(seconds= UT[j])
      #  print( UT[j], DT, AngleDelta[j], LenDelta[j], New_Method[j], Old_Method[j] )
    
    elif numpy.all(AngleDelta < ANGLE_LIMIT) and numpy.all(LenDelta < LENGTH_LIMIT):
      print("PASS: {:s}, Avg Angle = {:.6f} arcsec, Avg Len = {:.4f} um".format(Name, numpy.mean(AngleDelta*180/numpy.pi * 3600), numpy.mean(LenDelta*1e6) ))
    else:
      print("Issue encountered in the data for", Name)
        
  
  print()
  
  # Spherical comparison is slightly different
  # Values are R, Theta, Phi
  _New_car_to_sph = numpy.column_stack(_New_car_to_sph)
  _Old_car_to_sph = numpy.array       (_Old_car_to_sph)
  _Sph_Delta = numpy.abs(_New_car_to_sph - _Old_car_to_sph)
  _Alt_Delta = _Sph_Delta[:,0]
  _Lat_Delta = _Sph_Delta[:,1]
  _Lon_Delta = _Sph_Delta[:,2]
  #return _New_car_to_sph,_Old_car_to_sph
  
  if numpy.any( LENGTH_LIMIT < _Alt_Delta ) or numpy.any( ANGLE_LIMIT < _Lat_Delta) or numpy.any( ANGLE_LIMIT < _Lat_Delta ):
    print("FAIL: car_to_sph")
    for i in range(len( _New_car_to_sph )):
      if numpy.any( LENGTH_LIMIT < _Alt_Delta[i] ) or numpy.any( ANGLE_LIMIT < _Lat_Delta[i] ) or numpy.any( ANGLE_LIMIT < _Lat_Delta[i] ):
        print( UT[i])
  
  elif numpy.all( LENGTH_LIMIT > _Alt_Delta ) and numpy.all( ANGLE_LIMIT > _Lat_Delta) and numpy.all( ANGLE_LIMIT > _Lat_Delta ):
    print("PASS: car_to_sph , Pos Error {:.3f} um, Lat Err {:.6f} arcsec, Lon Error {:.6f} arcsec".format( numpy.mean(_Alt_Delta*1e6), numpy.mean( _Lat_Delta * 180/numpy.pi*3600), numpy.mean( _Lon_Delta * 180/numpy.pi*3600) ) ) 
  else:
    print("Issue parsing car_to_sph")
  
  
  _New_geod_to_geo = numpy.column_stack(_New_geod_to_geo)
  _Old_geod_to_geo = numpy.array       (_Old_geod_to_geo)
  _Geographic_Delta = numpy.abs(_New_geod_to_geo - _Old_geod_to_geo)
  _Alt_Delta = _Geographic_Delta[:,0]
  _Lat_Delta = _Geographic_Delta[:,1]
  
  if numpy.any( LENGTH_LIMIT < _Alt_Delta ) or numpy.any( ANGLE_LIMIT < _Lat_Delta):
    print("FAIL: geod_to_geo")
  elif not(numpy.any( LENGTH_LIMIT < _Alt_Delta ) or numpy.any( ANGLE_LIMIT < _Lat_Delta)):
    print("PASS: geod_to_geo, Alt Error {:.3f} um, Angle Err {:.6f} arcsec".format( numpy.mean(_Alt_Delta*1e6), numpy.mean( _Lat_Delta * 180/numpy.pi*3600)) ) 
  else:
    print("Issue parsing geo_to_geod")
  
  
    
  
  # And for geodetic angles: Just need to compare Lat/Altitue outputs.
  _New_geo_to_geod = numpy.column_stack(_New_geo_to_geod)
  _Old_geo_to_geod = numpy.array       (_Old_geo_to_geod)
  
  # And check the internal consistency
  _New_geo_to_geo  = numpy.column_stack( _New_geo_to_geo )
  
  _New_Old_Delta = numpy.abs(_New_geo_to_geod - _Old_geo_to_geod)
  _New_New_Delta = numpy.abs( _New_geo_to_geo - numpy.column_stack(( Geographic_Radius, Geographic_Colatitude)) )

  _New_Old_Alt_Delta = _New_Old_Delta[:,0]
  _New_Old_Lat_Delta = _New_Old_Delta[:,1]

  _New_New_Alt_Delta = _New_New_Delta[:,0]
  _New_New_Lat_Delta = _New_New_Delta[:,1]

  if numpy.any( LENGTH_LIMIT < _New_New_Alt_Delta ) or numpy.any( ANGLE_LIMIT < _New_New_Lat_Delta):
    print("FAIL: Internal GEO->Geod->GEO conversions")
  elif not(numpy.any( LENGTH_LIMIT < _Alt_Delta ) or numpy.any( ANGLE_LIMIT < _Lat_Delta)):
    print("PASS: Internal GEO->Geod->GEO conversions")
    print("                   Alt Error {:.3f} um, Angle Err {:.6f} arcsec".format( numpy.mean(_New_New_Alt_Delta*1e6), numpy.mean(_New_New_Lat_Delta*180/numpy.pi*3600) ))
  else:
    print("Issue parsing GEO->Geod->GEO")
  
  print()
  print("Due to a patch in GEOPACK's GEO->Geod conversion (early breaking),\n\tas well as a slight update to the WGS84 constants,the values vary on the order of a few hundred arcseconds.")
  print("Angle    Avg: {:.6f} arcsec, Max: {:.6f} arcsec".format( numpy.mean(_New_Old_Lat_Delta*180/numpy.pi*3600), numpy.max(_New_Old_Lat_Delta*180/numpy.pi*3600) ))
  print("Altitude Avg: {:.3f} meters, Max: {:.3f} meters".format( numpy.mean(_New_Old_Alt_Delta*1e3), numpy.max(_New_Old_Alt_Delta*1e3) ))
  

