import os
import sys
import datetime

import numpy
from   numpy.lib import recfunctions

from epop_ephemeris_library.rotation_libraries import lib_quaternion as lib_quat
from epop_ephemeris_library import database

# USER-CONTROL VARIABLES:
CFG_DISABLE_AUTO_FETCH = False # Rejects automatic fetching of IERS data constants. Allows for higher-level control.
CFG_DEFINITIVE_ONLY    = False # Only load defintive data points.
# By default, on any major call of this library, will fetch the entirety of the IERS dataset from the ephemeris_database.
# Set the above booleans to require higher-grained control of loaded data.

####
# Global Pathing:

# Global Shared Vars:
IERS_MET, IERS_X_Pole, IERS_Y_Pole, IERS_UT1_Minus_UTC, IERS_Delta_Psi,  IERS_Delta_Eplison, IERS_Length_of_Day, IERS_is_Definitive = [None] * 8
# MET,    Arcseconds,  Arcseconds,  Seconds,            Milliarcseconds, Milliarcseconds,    milliseconds,       boolean.

IERS_Structured_Array = None # The above, within a NP structured array for easier external reference.
#   | Column         | Type    |
#   |:------         |:----    |
#   | met            | float64 |
#   | X_Pole         | float64 |
#   | Y_Pole         | float64 |
#   | UT1_Minus_UTC  | float64 |
#   | Delta_Psi      | float64 |
#   | Delta_Eplison  | float64 |
#   | Length_of_Day  | float64 |
#   | Definitive     | boolean |
# Does NOT include the MJD or datetime, as raw dates are an inconsistent nightmare to work with.



##############################
# Global Convenience Functions:
_MET_Epoch = datetime.datetime(1968,5,24)   # Epoch of 1968-05-24 00:00:00 UTC
_MJD_Epoch = datetime.datetime(1858,11,17)  # Epoch of 1858-11-17 00:00:00 UTC

def _MET_to_ISO_8601_Str( MET ): # For use in debugging output.
  return (_MET_Epoch + datetime.timedelta(seconds=MET)).strftime("%Y-%m-%dT%H:%M:%S")
  
##############################
# Data Loading Functions.
def Load_IERS_Data( start_met:      float = None,
                    end_met:        float = None, 
                    IERS_filepath:  str   = None,
                    Defintive_Only: bool = CFG_DEFINITIVE_ONLY ):
  '''
  Loads the IERS rotation constants into local memory.
  Requires the 'IERS EOP 14 C04 (IAU1980)' publication data either be available on the ephemeris_database, or available locally.
  
  First tries to read the default filepath provided in <library>.IERS_Default_Path
  If that fails, download and read a copy of the IERS EOP 14 C04 (IAU1980) publication. Stored in a temporary file.
  If that fails, raise a verbose exception.
  
  start_met: Float. The earliest desired time to use. Rounds down to the prior day, as to allow for an interpolation point. If None, loads to the start of the dataset.
  end_met  : Float. The last desired time to use. Rounds up to the following day, as to allow for an interpolation point.  If None, loads to the end of the dataset.
    If either are given, checks that the given datarange is within the bounds requested.

  IERS_filepath: Filepath. If given, reads the given file for the IERS data. If None, pulls from the database (default).
  Defintive_Only: Boolean. If True, discards all Predicted data, truncating the IERS dataset to only the last Definitive value.
  '''
  if IERS_filepath is None: # Default: Pull from database.
    _database_fetch_IERS_Data( start_met, end_met, Defintive_Only )
  else: # Given a filepath to read the IERS data from.
    _read_read_IERS_Data( IERS_filepath, start_met, end_met, Defintive_Only )
  
  strMET = _MET_to_ISO_8601_Str # Truncate the function for readability.
  # After loading the data, make sure that the desired time-range is loaded into memory:
  if start_met is not None:
    if start_met < numpy.min( IERS_MET ):
      ErrorStr = "Pulled/Read IERS data starts {:s}, and requesting data starting at {:s}".format( strMET( numpy.min( IERS_MET ) ), strMET( start_met ) )
      raise ValueError( ErrorStr )
      
  if end_met is not None:
    if end_met < numpy.max( IERS_MET ):
      if Defintive_Only == False:
        ErrorStr = "Pulled/Read IERS data ends at {:s}, and requesting data stopping at {:s},".format( strMET( numpy.max( IERS_MET ) ), strMET( end_met ) )
      else:
        ErrorStr = "Pulled/Read IERS defintive data ends at {:s}, and requesting data stopping at {:s},".format( strMET( numpy.max( IERS_MET ) ), strMET( end_met ) )
      raise ValueError( ErrorStr )   
  
def _read_file_IERS_Data( 
            iers_file_path,
            start_met:  float = None,
            end_met:    float = None,
            definitive: bool = False ):
  '''
  Loads the IERS coefficient data into memory.
  Must follow the IERS EOP 14 C04 (IAU1980) file format.
  
  Returns:
    None
  
  Args:
    iers_file_path: A relative or absolute path to the given IERS file.
    start_met:      Starting MET to truncate the data. If None, does not truncate.
    end_met:        Ending MET to truncate the data. If None, does not truncate.
    definitive:     Loads only the Defintive dataset.
  '''
  # The file format used by IERS is... odd. Read the 'IERS_EOP_Documentation' file next to the EOP file for more details.
  
  FD = open( iers_file_path ).readlines()
  # Will raise a FileNotFoundError if could not read the file.
  
  # Don't bother with year/month, as has the Y2K %2d year wrap issue. Just read MJD.
  MJD                 = [ Line[  7: 15].strip() for Line in FD ] # Guaranteed populted for each line.
  # Read the Bulletin A values.
  Bulletin_A_X_Pole   = [ Line[ 18: 27].strip() for Line in FD ] # Published about 6-10 months ahead of current time.
  Bulletin_A_Y_Pole   = [ Line[ 37: 46].strip() for Line in FD ] # "
  Bulletin_A_UTC1_UTC = [ Line[ 58: 68].strip() for Line in FD ] # "
  Bulletin_A_LOD      = [ Line[ 79: 86].strip() for Line in FD ] # Published a few days behind scheule, treated as a 'definitive' value. Not always populated.
  Bulletin_A_dPsi     = [ Line[ 97:106].strip() for Line in FD ] # Published about 3 months ahead of current time.
  Bulletin_A_dEps     = [ Line[116:125].strip() for Line in FD ] # "
  # And then the Bulletin B solutions.
  Bulletin_B_X_Pole   = [ Line[134:144].strip() for Line in FD ] # Should be populated a most a month behind the current time.
  Bulletin_B_Y_Pole   = [ Line[144:154].strip() for Line in FD ] # "
  Bulletin_B_UTC1_UTC = [ Line[154:165].strip() for Line in FD ] # "
  Bulletin_B_dPsi     = [ Line[165:175].strip() for Line in FD ] # "
  Bulletin_B_dEps     = [ Line[175:185].strip() for Line in FD ] # "
  # Bulletin B does not publish the Length of day.
  
  # And then combine them, prefering the more accurate Bulletin-B solutions.
  X_Pole   = [ Bulletin_B_X_Pole  [i] if Bulletin_B_X_Pole  [i] != "" else Bulletin_A_X_Pole  [i] for i in range(len(FD)) ]
  Y_Pole   = [ Bulletin_B_Y_Pole  [i] if Bulletin_B_Y_Pole  [i] != "" else Bulletin_A_Y_Pole  [i] for i in range(len(FD)) ]
  UTC1_UTC = [ Bulletin_B_UTC1_UTC[i] if Bulletin_B_UTC1_UTC[i] != "" else Bulletin_A_UTC1_UTC[i] for i in range(len(FD)) ]
  LOD      = [ Bulletin_A_LOD     [i]                                                             for i in range(len(FD)) ] # No Bulletin-B LOD. Any given values are Definitive.
  dPsi     = [ Bulletin_B_dPsi    [i] if Bulletin_B_dPsi    [i] != "" else Bulletin_A_dPsi    [i] for i in range(len(FD)) ]
  dEps     = [ Bulletin_B_dEps    [i] if Bulletin_B_dEps    [i] != "" else Bulletin_A_dEps    [i] for i in range(len(FD)) ]
  
  # And then replace all empty strings with NaN for parsing by NP into floats.
  X_Pole   = [ Str if Str != "" else "NaN" for Str in X_Pole  ]
  Y_Pole   = [ Str if Str != "" else "NaN" for Str in Y_Pole  ]
  UTC1_UTC = [ Str if Str != "" else "NaN" for Str in UTC1_UTC]
  LOD      = [ Str if Str != "" else "NaN" for Str in LOD     ]
  dPsi     = [ Str if Str != "" else "NaN" for Str in dPsi    ]
  dEps     = [ Str if Str != "" else "NaN" for Str in dEps    ]
  
  # And for the Definitive mask, just take a bool mask any Bulletin-B list, and that's the boolean True mask!
  isDef = numpy.array([ Str != "" for Str in Bulletin_B_X_Pole  ])
  
  # From these, build the structured array.
  # Build the initial list of MET, append the remainder to it.
  MET  = numpy.array([(_MJD_Epoch - _MET_Epoch + datetime.timedelta(days=float(mjd))).total_seconds() for mjd in MJD ], dtype=float)
  data = numpy.array( MET, dtype=[('met','float64')] )
  
  _names = ['X_Pole','Y_Pole','UT1_Minus_UTC','Delta_Psi','Delta_Eplison','Length_of_Day','Definitive']
  _data  = [ X_Pole,  Y_Pole,  UTC1_UTC,       dPsi,       dEps,           LOD,            isDef      ]
  _types = ['float64'] * 6 + [bool]
  Structured_Data = recfunctions.append_fields( data, _names, _data, dtypes=_types, usemask=False)
  
  # And finally, truncate the dataset to the desired range.
  if start_met is not None:
    _Padded_Start = numpy.floor(( start_met - 10 ) / 86400 ) * 86400 # Anchor to the prior midnight of 10s before the start time to guarantee smooth interpolation.
    _Valid_Mask   = _Padded_Start <= Structured_Data['met']
    Structured_Data = Structured_Data[_Valid_Mask]
  if end_met is not None:
    _Padded_End = numpy.floor(( end_met + 10 )/ 86400 ) * 86400 + 86400 # Anchor to the following midnight of 10s after the saught end time to guarantee smooth interpolation.
    _Valid_Mask   = Structured_Data['met'] <= _Padded_End
    Structured_Data = Structured_Data[_Valid_Mask]
  if definitive:
    Structured_Data = Structured_Data[ Structured_Data['Definitive'] == True ]
  
  # And finally, assign it to the class variables.
  global IERS_MET, IERS_X_Pole, IERS_Y_Pole, IERS_UT1_Minus_UTC, IERS_Delta_Psi,  IERS_Delta_Eplison, IERS_Length_of_Day, IERS_is_Definitive
  global IERS_Structured_Array
  
  IERS_Structured_Array = Structured_Data
  IERS_MET           = Structured_Data['met']
  IERS_X_Pole        = Structured_Data['X_Pole']
  IERS_Y_Pole        = Structured_Data['Y_Pole']
  IERS_UT1_Minus_UTC = Structured_Data['UT1_Minus_UTC']
  IERS_Delta_Psi     = Structured_Data['Delta_Psi']
  IERS_Delta_Eplison = Structured_Data['Delta_Eplison']
  IERS_Length_of_Day = Structured_Data['Length_of_Day']
  IERS_is_Definitive = Structured_Data['Definitive']
  
def _database_fetch_IERS_Data(
             start_met:  int = None,
             end_met:    int = None,
             definitive: bool = False):
  '''
  Fetch IERS rotation constants from the ephemeris database.

  Arguments:
    start_met: Float, MET time. Minimum MET to fetch data for. If None, fetch to the start of the available data.
    end_met:   Float, MET time. Maximum MET to fetch data for. If None, fetch to the end of the available data.
    definitive: Boolean. Flag to only fetch definitive data. As of v1.2, will truncate the data, as Predicted and Definive data is merged as a single list.
    
  Return:
    None.
    
  Sets the IERS_* class variables, and populates the IERS_Structured_Array struct.
  '''
  
  Table_Name = 'iers_data'
  Database_Columns = ['met', 'pole_x', 'pole_y', 'ut1-utc',
    'delta_psi', 'delta_epsilon', 'length_of_day', 'definitive' ]
  
  # Wrap the column names in back-ticks for use in MySQL, and merge into a single string.
  Database_Fields = ['`' + f + '`' for f in Database_Columns]
  Database_Fields = ", ".join( Database_Fields )
  
  # Anchor the start/end values if needed.
  # Pad the given times by 10s, then anchor to the prior or following day as to guarantee at least an extra entry to smooth interpolation.
  if start_met is not None:
    _start_met = int( numpy.floor((start_met-10) / 86400) * 86400)
  if end_met is not None:
    _end_met   = int( numpy.ceil ((end_met+10) / 86400) * 86400)

  
  # Then construct the Where filter:
  if start_met is not None and end_met is not None:
    Time_Filter = '`met` BETWEEN {:d} AND {:d}'.format(_start_met,_end_met)
  elif start_met is not None:
    Time_Filter = '`met` >= {:d}'.format(_start_met)
  elif end_met is not None:
    Time_Filter = '`met` <= {:d}'.format(_end_met)
  else: # Both are None:
    Time_Filter = "" # Don't filter any solution.
    
  # And add a 'Definitive' filter if needed.
  if definitive:
    Def_Filter = '`definitive` is True'
  else:
    Def_Filter = ""
    
  # And then build the Where filter.
  if Time_Filter == "" and Def_Filter == "":
    Where_Statement = ""
  elif Time_Filter != "" and Def_Filter != "":
    Where_Statement = " WHERE {:s} AND {:s}".format( Time_Filter, Def_Filter )
  else: # And if only one is true, then just merge them.
    Where_Statement = " WHERE {:s}".format( Time_Filter + Def_Filter ) # Safe, as ""+"a" = "a"
  
  # Build the statement.
  statement = (  'SELECT {:s}'.format(Database_Fields)
               + ' FROM `cassiope_ephemeris`.`{:s}`'.format(Table_Name)
               + Where_Statement
               + ' ORDER BY `met` ASC' )

  data = database.Query( statement )
    
  dtype_list = [('met',           'float64'),
                ('X_Pole',        'float64'),
                ('Y_Pole',        'float64'),
                ('UT1_Minus_UTC', 'float64'),
                ('Delta_Psi',     'float64'),
                ('Delta_Eplison', 'float64'),
                ('Length_of_Day', 'float64'),
                ('Definitive',    'bool'   )]
  
  # Build the structured array:
  data = numpy.array(data, dtype=dtype_list)
               
  # And assign to the global variables!
  global IERS_MET, IERS_X_Pole, IERS_Y_Pole, IERS_UT1_Minus_UTC, IERS_Delta_Psi,  IERS_Delta_Eplison, IERS_Length_of_Day, IERS_is_Definitive
  global IERS_Structured_Array
  
  IERS_Structured_Array = data
  IERS_MET              = data['met']
  IERS_X_Pole           = data['X_Pole']
  IERS_Y_Pole           = data['Y_Pole']
  IERS_UT1_Minus_UTC    = data['UT1_Minus_UTC']
  IERS_Delta_Psi        = data['Delta_Psi']
  IERS_Delta_Eplison    = data['Delta_Eplison']
  IERS_Length_of_Day    = data['Length_of_Day']
  IERS_is_Definitive    = data['Definitive']
  
#####################
def _is_class_populated():
  '''
  A built-in sanity check to make sure the class variables are instantiated.
  Functionalized as to limit patch sizes when patching the class variables.
  '''
  return IERS_Structured_Array is not None
  
def Sanity_Check_IERS_Data( MET_List ):
  '''
  Given a MET list,make sure that all IERS times are within the given range.
  If IERS data not loaded, load using the default methods.
   
  Raises a ValueError should the requested list fall outside the constants file range.
  '''
  
  # If the data has yet to be pulled, and the automated pulling is not disabled, pull the data.
  if _is_class_populated() == False and CFG_DISABLE_AUTO_FETCH == False:
    Load_IERS_Data() # Load the global constants, if needed.
    
  elif _is_class_populated() == False and CFG_DISABLE_AUTO_FETCH == True:
    raise ValueError("lib_j2k_geo_rotation has automated IERS constant fetch disabled via CFG_DISABLE_AUTO_FETCH. Please manually populate with the class' Load_IERS_Data function.")
  
  # If data is successfully loaded and values are bound within desired range, return
  if numpy.min(IERS_MET) <= numpy.min(MET_List) and numpy.max(MET_List) <= numpy.max(IERS_MET):
    return
  else: # If data is not bound by the MET_List, raise exception.  
    raise ValueError("Requested MET outside the range of the loaded IERS_EOP data.")

##############################
# Coefficient-Fetching functions:

def Fetch_Delta_Symbols_MET( MET_List ):
  '''
  Returns the Delta_Psi and Delta_Epsilon values, linearly interploated.
  Values returned as two NP arrays, in RADIAN units.
  
  Args:
    MET_List: np array of floats
  Returns: 
    delta_Psi    : numpy.array, float. Radians.
    delta_Epsilon: numpy.array, float. Radians.
  '''
  if _is_class_populated() == False:
    Read_IERS_Data() # Load the globals, if needed.
  
  # Just apply a linear interpolation.
  delta_Psi     = numpy.interp( MET_List, IERS_MET, IERS_Delta_Psi )
  delta_Epsilon = numpy.interp( MET_List, IERS_MET, IERS_Delta_Eplison )
  
  # These are given by the IERS & stored as milliarcseconds, so translate to radians.
  delta_Psi     = delta_Psi     / 1000 / 60 / 60 * numpy.pi / 180 # mArcSec->ArcSec->ArcMin->Degree->Radian.
  delta_Epsilon = delta_Epsilon / 1000 / 60 / 60 * numpy.pi / 180
  
  return delta_Psi, delta_Epsilon

def Fetch_Pole_Wander_MET( MET_List ):
  '''
  Returns the X- and Y-Wander values, linearly interploated.
  Values returned as two NP arrays, in RADIAN units.
  
  Args:
    MET_List: np array of floats
  Returns: 
    X_Wander_Rad: numpy.array, float. Radians.
    Y_Wander_Rad: numpy.array, float. Radians.
  '''
  if _is_class_populated() == False:
    Read_IERS_Data() # Load the globals, if needed.
    
  # For now, just apply a linear interpolation.
  X_Wander_Arcsec = numpy.interp( MET_List, IERS_MET, IERS_X_Pole )
  Y_Wander_Arcsec = numpy.interp( MET_List, IERS_MET, IERS_Y_Pole )
  
  # These are given and stored as arcseconds, so translate to radians.
  X_Wander_Rad = X_Wander_Arcsec / 3600 * numpy.pi / 180
  Y_Wander_Rad = Y_Wander_Arcsec / 3600 * numpy.pi / 180
  
  return X_Wander_Rad, Y_Wander_Rad

def Fetch_UTC_UT1_Correction_MET( MET_List):
  '''
  Returns the UTC-UT1 values, linearly interploated.
  Values returned as an NP arrays, in seconds, expressed as floats.
  
  Args:
    MET_List: np array of floats
  Returns: 
    UT1_Minus_UTC_Seconds: numpy.array, float. Seconds.
  '''
  if _is_class_populated() == False:
    Read_IERS_Data() # Load the globals, if needed.
    
  # Just apply a linear interpolation.
  UT1_Minus_UTC_Seconds = numpy.interp( MET_List, IERS_MET, IERS_UT1_Minus_UTC )
  # Can just return. They're floating-point seconds.
  return UT1_Minus_UTC_Seconds

##############################
# Time Calculations.

def Calculate_Julian_Day( MET_List ): 
  '''
  Returns the Julian Day numbers of the given MET list.
  The given timestamps do not have to be UTC-keyed, like MET.
  (Eg: TDT, TDT, TAI)
  
  TLDR: Returns solar days since Noon Jan 1 4713 BC
  J2K Epoch = 2000-01-01 12:00 UTC = MET 997444800 = 2,451,545.0 JD
  MET Epoch = 1968-05-24 00:00 UTC = MET 0         = 2,440,000.5 JD
      
  Args:
    MET_List: np array of floats
  Returns:
    julian_day_list: np array of floats
  '''

  # JD at MET 0 == 2440000.5
  # Thus just add the number of days (MET/86400) to the MET(0) JD value.
  julian_day_list = 2440000.5 + MET_List / 86400
  return julian_day_list

def Calculate_TDB( MET_UTC_List ):
  '''
  Derives the Barycentric Dynamical Time from UTC-MET. 
  TDB is the most accurate Time system for astronomical systems, as it accounts for t
  
  Returns a NP array of float timestamps, expressed as seconds since the MET epoch.
     
  Args:
    MET_UTC_List: np array of floats. Seconds since the MET epoch. Must be UTC-keyed.
  Returns:
    MET_TDB_List: np array of floats. Seconds since the MET epoch. TDB-keyed
  '''
  # TDB = TDT + 0.001658(sec) * sin(g+0.10167 * sin(g))
  # g = 2*pi*( 357.528* + 35999.50* * T) / 360*
  # T = ( JD(TDT) - 3451545 ) / 36525)    [Can use JD(TDB) to invert without significant error]
  
  # TDT = TAI + 32.184s
  # TAI = UTC + Leap_Seconds + 10

  def Get_Leap_Seconds( MET_UTC_List ):
    # Only considers leap-seconds after launch.
    if numpy.min(MET_UTC_List) <= 1431302400: raise ValueError("Using MET before Launch.") # Sanity check.
    
    Leap_Seconds = numpy.full(MET_UTC_List.shape, 25) # At launch, UTC-TAI was 35 (25+10)
    Leap_Seconds[ MET_UTC_List >= 1518048000 ] += 1 # 2016-07-01
    Leap_Seconds[ MET_UTC_List >= 1533945600 ] += 1 # 2017-01-01
    return Leap_Seconds    
  
  MET = MET_UTC_List
  TAI = MET + Get_Leap_Seconds(MET) + 10
  TDT = TAI + 32.184
  T   = ( Calculate_Julian_Day(TDT) - 2451545) / 36525
  
  g    = 2*numpy.pi * (357.528 + 35999.050 * T) / 360 # Rad * Deg/Deg = Rad
  _sec = 0.001658 * numpy.sin( g + 0.10167 * numpy.sin(g))
  
  MET_TDB_List  = TDT + _sec
  return MET_TDB_List
  
def Calculate_Julian_Centures_Since_J2K( MET ):
  '''
  Given MET keyed to UTC, return Julian Centuries since the J2K epoch in the TDB timekey.
  
  Julian Century = 36525 Julian days. (100x365 days)
  J2K epoch:
    2000-01-01 12:00:00.000 TT = 2451545.0 JD TT (Terrestrial Time)
    2000-01-01 11:59:27.816 TAI
    2000-01-01 11:58:55.816 UTC
  
  Terrestrial Time           = TAI + 32.184
  Barycentric Dynamical Time = TAI + 32.184 +/- a few milliseconds
  
  Calculations below use TDB, as following documentation.
  
  Args:
    MET: np array of floats. Seconds since the MET epoch. Must be UTC-keyed.
  Returns:
    JC:  np array of floats. Julian centuries since the J2k epoch. TDB-keyed
  '''
  TDB = Calculate_TDB       ( MET )
  JD  = Calculate_Julian_Day( TDB )
  return (JD - 2451545.0) / 36525
  
def Calculate_Seconds_Into_Day( MET_UT1 ):
  '''
  Given MET keyed to UT1, return the current rotation fraction of the day.
  As MET bound on the midnight, any MET_UT1 % 86400 = 0, which is midnight.
  Just returns MET_UT1 % 86400
  
  Args:
    MET_UT1: np array of floats. Seconds since the MET epoch. Must be UT1-keyed.
  Returns:
    Seconds_into_day: np array of floats.
  '''
  return MET_UT1 % 86400
      
##############################
# Quaternion Calculations.

# The nominal R1/R2/R3 Roll/Pitch/Yaw rotation matricies.
# Returns Quaternions for each Radian entry given.
# Note that while given Radians, for consistancy, must translate to Degrees for the Quaternion Library.
# Also, the R1/R2/R3 values are sign flipped vs the Quaternion Library's requirement.
# These are kept sign-flipped in input as to follow the documentation for the R1/2/3 rotation matricies.
# Thus the Rad->Deg rotation in the function also flips the sign.

def R1_Quat( Rad ): # Roll-Quaternions
  _Roll  = Rad * 180 / numpy.pi
  _Zeros = numpy.full(Rad.shape, 0)
  YPR    = numpy.column_stack(( _Zeros, _Zeros, _Roll ))
  return lib_quat.YPR_to_Quaternion(YPR)
  
def R2_Quat( Rad ): # Pitch-Quaternions
  _Pitch = Rad * 180 / numpy.pi
  _Zeros = numpy.full(Rad.shape, 0)
  YPR    = numpy.column_stack(( _Zeros, _Pitch, _Zeros ))
  return lib_quat.YPR_to_Quaternion(YPR)
  
def R3_Quat( Rad ): # Yaw-Quaternions
  _Yaw   = Rad * 180 / numpy.pi
  _Zeros = numpy.full(Rad.shape, 0)
  YPR    = numpy.column_stack(( _Yaw, _Zeros, _Zeros ))
  return lib_quat.YPR_to_Quaternion(YPR)

#########################################
# Frame Tranlsation Calculations.

# Precession: P
# Returns a Quaternion List with the Processing correction for each MET timestamp.
# Not Depenant on any external values.
def Calculate_Precession_Correction( MET ):
  # Calculate Julian Centuries since J2K. (Float)
  T = Calculate_Julian_Centures_Since_J2K( MET )
  
  z = (2306.2181/3600 * T + 1.09468/3600 * T**2 + 0.018203/3600 * T**3) * numpy.pi/180 # Radians.
  v = (2004.3109/3600 * T - 0.42665/3600 * T**2 - 0.041833/3600 * T**3) * numpy.pi/180
  h = (2306.2181/3600 * T + 0.30188/3600 * T**2 + 0.017998/3600 * T**3) * numpy.pi/180
  # h = Zeta (Greek).
  
  # Returns Arrays of Quaternion.
  qRot_1 = R3_Quat( -z )
  qRot_2 = R2_Quat(  v )
  qRot_3 = R3_Quat( -h )

  return lib_quat._qChain_Multiply((qRot_1, qRot_2, qRot_3)) # Returns Quaternions

# Nutation: N
# Return a Quaternion List for each MET-Entry given.
# Dependant on the IERS data file for the Delta Psi and Epison values.
def Calculate_Nutation_Correction( MET ):
  # Calculate Julian Centuries since J2K. (Float)
  T = Calculate_Julian_Centures_Since_J2K( MET )

  Epsilon = (23 + 26/60 + 21.488/3600) - 46.8150/3600 * T - 0.00059/3600 * T**2 + 0.001813/3600 * T**3
  Epsilon = Epsilon * numpy.pi / 180 # Translate to Radians.
  
  delta_Psi, delta_Epsilon = Fetch_Delta_Symbols_MET( MET )    
    
  qRot_1 = R1_Quat( -(Epsilon + delta_Epsilon))
  qRot_2 = R3_Quat( -delta_Psi )
  qRot_3 = R1_Quat(  Epsilon )
  
  return  lib_quat._qChain_Multiply((qRot_1, qRot_2, qRot_3)) # Returns Quaternions
        
# Diurnal Matrix: R_S
# Returns a Single-Axis quaternion, along +Z. 
def Calculate_Diurnal_Correction( MET , Nutation_Quaternions = None ):
  # Derive the UT1 time. It's still expressed as MET, to keep that in it's name.
  UT1_MET = MET + Fetch_UTC_UT1_Correction_MET( MET )
  
  # Then calculate the Julian centures since the J2K epoch in UT1 (By definition 2000-01-01 12h UT1 (Noon))
  # Do not use Calc_Jul_Cent, as it translates UTC to TDB first.
  T_u = (Calculate_Julian_Day( UT1_MET ) - 2451545.0 ) / 36525.0
  
  # The Alpha-Epsilon is extracted from the Nutation Matrix:
  # If not given, calculate it.
  if Nutation_Quaternions is None: 
    Nutation_Quaternions = Calculate_Nutation_Correction( MET )
  
  # Extract the Rotation matrix, use to extract the hour-angle between the True Equinox of Date and the Mean Equinox of date.
  _Nutation_Rotation_Matrix = lib_quat.Quaternion_to_Rotation_Matrix( Nutation_Quaternions )
  Alpha_Epsilon = numpy.arctan( _Nutation_Rotation_Matrix[:,0,1] / _Nutation_Rotation_Matrix[:,0,0] ) # Radians

  # Then calculate the theta_G_0 value:
  # Expressed as Terrestrial Time, as 86400s=360*. So 1h = 15*, 1m = 1/4* = 15', 1s = 1/240* == 15"
  # Express theta_G_0 in -SECONDS- (The Time unit, == 15 arcseconds)
  theta_G_0 = 6*60*60 + 41*60 + 50.54841 + 8640184.812866 * T_u + 0.093104 * T_u**2 - 6.2e-6 * T_u**3 # in "Seconds"
  theta_G_0 = theta_G_0 / (24*60*60) * 360 * numpy.pi / 180 # "Seconds" -> Rotations -> -> Degrees -> Radians.
  
  # Then, apply the UT1 observation time (The TIME component of the UT1 timestamp)
  _UT1_Seconds_Into_Day = Calculate_Seconds_Into_Day( UT1_MET ) # Bascially just returns the modulo of 86400.
  UT1_Angle_Into_Day    = _UT1_Seconds_Into_Day / (24*60*60) * 360 * numpy.pi / 180  # "Seconds" -> Degrees -> Radians
  
  # Change angle range from {0:360} to {-180:180}
  UT1_Angle_Into_Day[ UT1_Angle_Into_Day >= numpy.pi ] -= 2*numpy.pi

  # The given multiplier-float below has caused nothing but headaches.
  # The replacement 1.0 decreases error, and increases consistancy. No clue as to why.
  #theta_G   = 1.002737909350795 * UT1_Angle_Into_Day + theta_G_0
  theta_G   = 1.0 * UT1_Angle_Into_Day + theta_G_0
  #theta_G   = (2-1.002737909350795) * UT1_Angle_Into_Day + theta_G_0
  
  Theta_G = theta_G + Alpha_Epsilon # Theta = Capital theta. Can't exactly use greek symbols in python.

  return R3_Quat( Theta_G )

# Polar Motion: R_M
def Calculate_Polar_Motion_Correction( MET ):
  X_Wander, Y_Wander = Fetch_Pole_Wander_MET( MET ) # Radians
  
  qX_Wander = R2_Quat( -X_Wander )
  qY_Wander = R1_Quat( -Y_Wander )
  
  return lib_quat._qMultiply(qX_Wander, qY_Wander)
  
#######################################################
#######################################################
  
def Build_ICRF_to_ITRF_Quaternions( MET_Timestamps ):
  '''
  Given a list of timestamps in MET, 
  Return the J2K->GEO == ICRF->ITRF quaternions at those timestamps.
  Note that these are RIGHT-HANDED quaternions.
  s.t. vITRF = qICRF->ITRF* . vICRF . qICRF->ITRF
  '''
  # r_CEP  =  N  .  P  . r_ICRF
  # r_ITRF = R_M . R_S . r_CEP
  # Thus
  # r_ITRF = R_M . R_S . ( N .  P . r_ICRF )
  # r_ITRF = R_M . R_S . N . P . r_ICRF
  # Thus
  # rITRF<-ICRF = R_M . R_S . N . P  # Using Left-Handed Rotation-matrix chaining.
  # But as building and returning Quaternions, rotation is RIGHT-handed.
  # s.t.
  # qICRF->ITRF = qP . qN . qR_S . qR_M
  # vITRF = qICRF->ITRF* . vICRF . qICRF->ITRF
  
  MET_Timestamps = numpy.array(MET_Timestamps, dtype=float)
  
  # Apply sanity checks first:
  Sanity_Check_IERS_Data( MET_Timestamps )
  
  P   = Calculate_Precession_Correction  ( MET_Timestamps ) # ICRF->P
  N   = Calculate_Nutation_Correction    ( MET_Timestamps ) # P   ->N
  R_S = Calculate_Diurnal_Correction     ( MET_Timestamps ) # N   ->R_S
  R_M = Calculate_Polar_Motion_Correction( MET_Timestamps ) # R_S ->ITRF

  qJ2K_to_GEO = qICRF_to_ITRF = lib_quat._qChain_Multiply(( P, N, R_S, R_M ))
  
  return qICRF_to_ITRF

# Sanity for those who don't want to deal with quaternions.
# Returns a Nx3x3 NP array, where N is the given MET array length.
def Build_ITRF_from_ICRF_Rotation_Matricies( MET_Timestamps ):
  '''
  Given a list of timestamps in MET, 
  Return the rotation matrix of GEO<-J2K == ITRF<-ICRF at those timestamps.
  Note that these are LEFT-HANDED rotation matricies.
  s.t. vITRF = rITRF<-ICRF . vICRF
  '''
  qICRF_to_ITRF   = Build_ICRF_to_ITRF_Quaternions( MET_Timestamps )
  rITRF_from_ICRF = lib_quat.Quaternion_to_Rotation_Matrix( qICRF_to_ITRF ) # Change 'to/from' notation to re-inforce handedness change.
  return rITRF_from_ICRF
  
############################################################
############################################################

# Sane transforms
# To rotate a Nx3x3 matrix-list by a Nx3 vector list in Numpy is a nuciance.
# Here the A->B quaternion is fetched, and the lib_quaternion.Vector_Rotate function used in it's stead
def GEO_to_J2K( MET, geo_x, geo_y, geo_z ): 
  '''
  Given a list of times & GEO coordinates, returns the equivalent J2K coordinates.
  Note: KNOWN ERROR. <50" at max, usually <10".
  
  Args:
    MET:   nd.array of floats. MET timestamps.
    geo_x: nd.array of floats. X-coordinates in GEO. (Arbitrary units)
    geo_y: nd.array of floats. Y-coordinates in GEO. (Arbitrary units)
    geo_z: nd.array of floats. Z-coordinates in GEO. (Arbitrary units)
    
  Returns:
    j2k_x: nd.array of floats. X-coordinates in J2K. (Same units as given in args)
    j2k_y: nd.array of floats. Y-coordinates in J2K. (Same units as given in args)
    j2k_z: nd.array of floats. Z-coordinates in J2K. (Same units as given in args)
  '''
  qJ2K_to_GEO = Build_ICRF_to_ITRF_Quaternions( MET ) # The rotation of J2K->GEO, qGEO<-J2k
  qGEO_to_J2K = lib_quat.Conjugate( qJ2K_to_GEO )     # Conjugate = Quaternion inversion/transpose.
  
  # Vector_Rotate takes an Nx3 matrix, so stack the Nx1 position columns:
  geo_vector = numpy.column_stack(( geo_x, geo_y, geo_z ))
  j2k_vector = lib_quat.Vector_Rotate( qGEO_to_J2K, geo_vector )
  # And then split to the individual columns.
  j2k_x, j2k_y, j2k_z = j2k_vector.T
  
  return j2k_x, j2k_y, j2k_z
 
def J2K_to_GEO( MET, j2k_x, j2k_y, j2k_z ):
  '''
  Given a list of times & J2K coordinates, returns the equivalent GEO coordinates.
  Note: KNOWN ERROR. <50" at max, usually <10".
  
  Args:
    MET:   nd.array of floats. MET timestamps.
    j2k_x: nd.array of floats. X-coordinates in J2K. (Arbitrary units)
    j2k_y: nd.array of floats. Y-coordinates in J2K. (Arbitrary units)
    j2k_z: nd.array of floats. Z-coordinates in J2K. (Arbitrary units)
    
  Returns:
    geo_x: nd.array of floats. X-coordinates in GEO. (Same units as given in args)
    geo_y: nd.array of floats. Y-coordinates in GEO. (Same units as given in args)
    geo_z: nd.array of floats. Z-coordinates in GEO. (Same units as given in args)
  '''
  qJ2K_to_GEO = Build_ICRF_to_ITRF_Quaternions( MET )
    
  # Vector_Rotate takes an Nx3 matrix, so stack the Nx1 position columns:
  j2k_vector = numpy.column_stack(( j2k_x, j2k_y, j2k_z ))
  geo_vector = lib_quat.Vector_Rotate( qJ2K_to_GEO, j2k_vector )
  # And then split to the individual columns.
  geo_x, geo_y, geo_z = geo_vector.T
  
  return geo_x, geo_y, geo_z

 
############################################################
############################################################

def _Test_Known_Ephemeris(J2K_Ephemeris, GEO_Ephemeris):
  '''
  Test harness for the J2K<->GEO rotation matricies.
  Given GEO & J2K Ephemeris data over an equal region,
  Derive the rotation, check accuracy.
  
  Args
    J2K_Ephemeris - A structured array from ephemeris.get_j2k_ephemeris()
    GEO_Ephemeris - A structured array from ephemeris.get_geo_ephemeris()
  Returns
    None - Prints relevant data.
  '''
  # Make sure data is equivocable:
  if len(J2K_Ephemeris) != len(GEO_Ephemeris):
    raise ValueError("Given unequal lengths of ephemeris data!")
  if not numpy.all( J2K_Ephemeris['met'] == GEO_Ephemeris['met'] ):
    raise ValueError("Given timestamps lists are not equal!")
  
  # Extract only the MET & Positional data.
  MET     = J2K_Ephemeris['met']
  J2K_Pos = numpy.column_stack(( J2K_Ephemeris['j2k_x'], J2K_Ephemeris['j2k_y'], J2K_Ephemeris['j2k_z'] ))
  GEO_Pos = numpy.column_stack(( GEO_Ephemeris['geo_x'], GEO_Ephemeris['geo_y'], GEO_Ephemeris['geo_z'] ))
  
  # Fetch the rotation matricies.
  qGEO_from_J2K = Build_ICRF_to_ITRF_Quaternions( MET )
  rGEO_from_J2K = Build_ITRF_from_ICRF_Rotation_Matricies( MET )

  # Rotate the J2K into the GEO frame:
  GEO_Pos_Derived = lib_quat.Vector_Rotate( qGEO_from_J2K, J2K_Pos )
  
  # Calculate the total position error:
  Position_Error = numpy.sqrt(numpy.sum((GEO_Pos - GEO_Pos_Derived)**2, axis=1))
  
  # Derive the Normalized GEO vectors to simplify angle-calculation
  Norm_GEO, Norm_GEO_Derived = lib_quat.Normalize(GEO_Pos), lib_quat.Normalize(GEO_Pos_Derived)
  # Angle = arccos( v1.v2 )
  Angle_Error = numpy.arccos( [numpy.dot(Norm_GEO[i], Norm_GEO_Derived[i]) for i in range(len(Norm_GEO))] ) * 180/numpy.pi
  
  print("Stats")
  print("Avg Divergence: {:.2f} meters, {:.2f} arcseconds".format( numpy.mean(Position_Error), numpy.mean( Angle_Error) * 3600 ))
  print("Min Divergence: {:.2f} meters, {:.2f} arcseconds".format( numpy.min (Position_Error), numpy.min ( Angle_Error) * 3600 ))
  print("Max Divergence: {:.2f} meters, {:.2f} arcseconds".format( numpy.max (Position_Error), numpy.max ( Angle_Error) * 3600 ))

  

  
if __name__ == "__main__":
  print("! lib_j2k_geo_rotation has no CLI interface. Use by importing.")
  print("! Primary interface is Build_ITRF_from_ICRF_Quaternions and Build_ITRF_from_ICRF_Rotation_Matricies")
  print("! Exiting")
  sys.exit(1)








